from __future__ import unicode_literals
import io
import os
import tempfile
import json

import pytest

from .. import util


@pytest.fixture
def tmp_path():
    return tempfile.mkdtemp()


@pytest.fixture
def tmp_path2():
    return tmp_path()


@pytest.fixture(scope='module')
def sample_package_path(version='1.0.0'):
    path = tempfile.mkdtemp()

    with io.open(os.path.join(path, 'package.json'), 'wb') as f:
        f.write(util.json_dump({
            "name": "test",
            "description": "test package",
            "include": [("data", "*")],
            "version": version,
            "license": "public domain",
            "compatibility": {"test": None}
        }))
    data_path = os.path.join(path, 'data')
    os.mkdir(data_path)
    with io.open(os.path.join(data_path, 'xyz.model'), 'w') as f:
        for i in range(1):
            f.write(('%s' % i) * 1024)
    with open(os.path.join(data_path, 'xyz.json'), 'w') as f:
        json.dump({'test': True}, f)
    return path


@pytest.fixture(scope='module')
def sample_package_path2():
    return sample_package_path('2.0.0')


def pytest_addoption(parser):
    parser.addoption("--remote", action="store_true",
        help="include tests that require internet connectivity")


def pytest_runtest_setup(item):
    for opt in ['remote']:
        if opt in item.keywords and not item.config.getoption("--%s" % opt):
            pytest.skip("need --%s option to run" % opt)



# pylint: disable=C0330
import argparse
import logging

from . import default
from . import install, build, remove, search, find, upload, update, files, purge


def set_log_level(args):
    level = getattr(logging, args.log_level.upper())
    logging.basicConfig(level=level)


def add_build_parser(subparsers):
    parser = subparsers.add_parser('build',
        help='build package from package.json')
    parser.add_argument('package_path',
        default=default.build_package_path,
        nargs='?',
        help='package.json directory')
    parser.add_argument('archive_path',
        nargs='?',
        help='archive path')

    def run(args):
        set_log_level(args)
        build(package_path=args.package_path,
              archive_path=args.archive_path)

    parser.set_defaults(run=run)


def add_install_parser(subparsers):
    parser = subparsers.add_parser('install',
        help='install package from repository or filesystem')
    parser.add_argument('package_name',
        help='package name or path')

    def run(args):
        set_log_level(args)
        install(app_name=args.name,
                app_version=args.version,
                package_name=args.package_name,
                data_path=args.data_path,
                repository_url=args.repository_url)

    parser.set_defaults(run=run)


def add_remove_parser(subparsers):
    parser = subparsers.add_parser('remove',
        help='remove installed package')
    parser.add_argument('package_string',
        help='package string')

    def run(args):
        set_log_level(args)
        remove(app_name=args.name,
               app_version=args.version,
               package_string=args.package_string,
               data_path=args.data_path)

    parser.set_defaults(run=run)


def add_find_parser(subparsers):
    parser = subparsers.add_parser('find',
        help='find installed packages')
    parser.add_argument('package_string',
        default=default.find_package_string,
        nargs="?",
        help='package string')
    parser.add_argument('--meta',
        default=default.find_meta,
        action='store_true',
        help='show package meta data')
    parser.add_argument('--cache',
        default=default.find_cache,
        action='store_true',
        help='find cached instead of installed packages')

    def run(args):
        set_log_level(args)
        find(app_name=args.name,
             app_version=args.version,
             package_string=args.package_string,
             meta=args.meta,
             cache=args.cache,
             data_path=args.data_path)

    parser.set_defaults(run=run)


def add_search_parser(subparsers):
    parser = subparsers.add_parser('search',
        help='search installable packages on repository')
    parser.add_argument('search_string',
        nargs="?",
        help='search string')

    def run(args):
        set_log_level(args)
        search(app_name=args.name,
               app_version=args.version,
               search_string=args.search_string,
               data_path=args.data_path,
               repository_url=args.repository_url)

    parser.set_defaults(run=run)


def add_upload_parser(subparsers):
    parser = subparsers.add_parser('upload',
        help='upload package')
    parser.add_argument('package_path',
        help='package path')

    def run(args):
        set_log_level(args)
        upload(app_name=args.name,
               app_version=args.version,
               package_path=args.package_path,
               data_path=args.data_path,
               repository_url=args.repository_url)

    parser.set_defaults(run=run)


def add_update_parser(subparsers):
    parser = subparsers.add_parser('update',
        help='update package cache')

    def run(args):
        set_log_level(args)
        update(app_name=args.name,
               app_version=args.version,
               data_path=args.data_path,
               repository_url=args.repository_url)

    parser.set_defaults(run=run)


def add_files_parser(subparsers):
    parser = subparsers.add_parser('files',
        help='displays package files')
    parser.add_argument('package_string',
        help='package string')

    def run(args):
        set_log_level(args)
        files(app_name=args.name,
              app_version=args.version,
              package_string=args.package_string,
              data_path=args.data_path)

    parser.set_defaults(run=run)


def add_purge_parser(subparsers):
    parser = subparsers.add_parser('purge',
        help='purges downloaded data')
    parser.add_argument('--cache',
        default=False,
        action='store_true',
        help='purge cache (cached packages)')
    parser.add_argument('--pool',
        default=False,
        action='store_true',
        help='purge pool (installed packages)')

    def run(args):
        set_log_level(args)
        purge(app_name=args.name,
              app_version=args.version,
              cache=args.cache,
              pool=args.pool,
              data_path=args.data_path)

    parser.set_defaults(run=run)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--name',
        help='app name')
    parser.add_argument('--version',
        help='app version')
    parser.add_argument('--data-path',
        help='data storage path')
    parser.add_argument('--repository-url',
        help='package repository path')
    parser.add_argument('--log-level',
        default='info',
        help='log level (default: info)')

    subparsers = parser.add_subparsers()
    add_build_parser(subparsers)
    add_install_parser(subparsers)
    add_remove_parser(subparsers)
    add_find_parser(subparsers)
    add_search_parser(subparsers)
    add_upload_parser(subparsers)
    add_update_parser(subparsers)
    add_files_parser(subparsers)
    add_purge_parser(subparsers)

    return parser



import os
import io

from . import util
from . import default
from .package_stub import PackageStub
from .archive_reader import ArchiveReader


class Archive(PackageStub):
    def __init__(self, path):
        self.archive = ArchiveReader(path)
        defaults = self.archive.get_member('package')
        super(Archive, self).__init__(defaults)

        self.path = path

    @property
    def manifest(self):
        return self.archive.get_member('manifest')

    def fileobjs(self):
        return {
            os.path.join(self.ident, default.META_FILENAME):
                io.BytesIO(util.json_dump(self.archive.meta)),
            os.path.join(self.ident, default.ARCHIVE_FILENAME):
                self.archive.archive
        }



import os
import json

import pytest

from ..recipe import Recipe
from ..package import Package, NotIncludedException
from ..archive import Archive
from ..pool import Pool
from ..dir_package import DirPackage


def test_build_and_check_archive(tmp_path, sample_package_path):
    recipe = Recipe(sample_package_path)
    archive1 = recipe.build(tmp_path)

    assert os.path.isfile(archive1.path)

    archive2 = Archive(archive1.path)

    for key in Archive.keys:
        assert getattr(archive1, key) == getattr(archive2, key)


def test_package_file_load(tmp_path, tmp_path2, sample_package_path):
    recipe = Recipe(sample_package_path)
    archive = Archive(recipe.build(tmp_path).path)
    pool = Pool('test', '1.0.0', tmp_path2)
    path = pool.install(archive)
    package = Package(path=path)

    assert package.path == path

    assert package.has_file('data', 'xyz.json')
    assert not package.has_file('data', 'foo')

    with package.open(['data', 'xyz.json']) as f:
        assert json.load(f) == {'test': True}

    with package.open(['data', 'xyz.json'], mode='rb', encoding=None) as f:
        data = f.read()
        assert data == json.dumps({'test': True}).encode('ascii')
        assert data == b'{"test": true}'

    with package.open(['data', 'foo'], default=0) as f:
        assert f == 0

    with package.open(['data', 'foo'], default=None) as f:
        assert f is None

    with pytest.raises(Exception):
        with package.open(['data', 'foo'], default=Exception) as f:
           pass


def test_package_file_path(tmp_path, tmp_path2, sample_package_path):
    recipe = Recipe(sample_package_path)
    archive = Archive(recipe.build(tmp_path).path)
    pool = Pool('test', '1.0.0', tmp_path2)
    path = pool.install(archive)
    package = Package(path=path)

    assert package.path == path

    assert package.has_file('data', 'xyz.model')
    assert package.file_path('data', 'xyz.model') == os.path.join(package.path, 'data', 'xyz.model')
    assert package.dir_path('data') == os.path.join(package.path, 'data')

    assert not package.has_file('data')
    assert not package.has_file('data', 'model')
    with pytest.raises(NotIncludedException):
        assert package.file_path('data', 'model')


def test_file_path_same_build_directory(tmp_path, sample_package_path):
    recipe = Recipe(sample_package_path)
    archive = Archive(recipe.build(sample_package_path).path)
    pool = Pool('test', '1.0.0', tmp_path)
    package = Package(path=pool.install(archive))

    assert package.has_file('data', 'xyz.model')
    assert package.file_path('data', 'xyz.model') == os.path.join(package.path, 'data', 'xyz.model')
    assert package.dir_path('data') == os.path.join(package.path, 'data')

    assert not package.has_file('data')
    assert not package.has_file('data', 'model')
    with pytest.raises(NotIncludedException):
        assert package.file_path('data', 'model')


@pytest.mark.xfail
def test_new_archive_files(tmp_path, sample_package_path):
    recipe = Recipe(sample_package_path)
    archive = recipe.build(tmp_path)

    assert archive.manifest
    assert set([m['path'] for m in archive.manifest]) == \
           set([('data', 'xyz.model'), ('data/xyz.json')])


def test_archive_files(tmp_path, sample_package_path):
    recipe = Recipe(sample_package_path)
    new_archive = recipe.build(tmp_path)
    archive = Archive(new_archive.path)

    assert archive.manifest
    assert set([tuple(m['path']) for m in archive.manifest]) == \
           set([('data', 'xyz.model'), ('data', 'xyz.json')])


def test_dir_package(sample_package_path):
    package = DirPackage(sample_package_path)

    assert package.path == sample_package_path

    assert package.has_file('data', 'xyz.json')
    assert not package.has_file('data', 'foo')

    assert package.dir_path('data') == os.path.join(sample_package_path, 'data')
    assert package.file_path('data', 'xyz.model') == os.path.join(sample_package_path, 'data', 'xyz.model')

    assert package.dir_path('data') == os.path.join(sample_package_path, 'data')

    with package.open(['data', 'xyz.json']) as f:
        assert json.load(f) == {'test': True}

    with package.open(['data', 'xyz.json'], mode='rb', encoding=None) as f:
        assert f.read() == json.dumps({'test': True}).encode('ascii') == b'{"test": true}'



import os
import logging
import shutil

from . import default
from . import util
from .package_stub import PackageStub


class PackageNotFoundException(Exception): pass
class CompatiblePackageNotFoundException(Exception): pass
class InvalidDataPathException(Exception): pass


class PackageList(object):

    package_class = PackageStub

    def __init__(self, app_name, app_version, path, **kwargs):
        super(PackageList, self).__init__()

        self.logger = logging.getLogger(__name__)

        self.app_name = app_name
        self.app_version = app_version

        self.path = path
        self.data_path = kwargs.get('data_path') or path

        if not self.data_path:
            raise InvalidDataPathException(self.data_path)

        self.load()

    def packages(self):
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        for path in os.listdir(self.path):
            if path.endswith('.tmp'):
                continue

            meta_path = os.path.join(self.path, path, default.META_FILENAME)
            if not os.path.isfile(meta_path):
                continue

            yield self.__class__.package_class(path=os.path.join(self.path, path))

    def load(self):
        self._packages = {}
        for package in self.packages():
            self._packages[package.ident] = package

    def get(self, package_string):
        candidates = self.find(util.split_package_string(package_string)[0])
        if not candidates:
            raise PackageNotFoundException(package_string)

        candidates = sorted(self.find(package_string))
        if not candidates:
            raise CompatiblePackageNotFoundException(package_string)
        return candidates[-1]

    def find(self, package_string=None):
        res = []
        for package in self._packages.values():
            name, constraint = util.split_package_string(package_string)
            if not name or name == package.name and util.constraint_match(constraint, package.version):
                res.append(package)
        return res

    def purge(self):
        self.logger.info('purging %s', self.__class__.__name__)
        for package in self.find():
            self.remove(package)

    def remove(self, package):
        if not os.path.isdir(package.path):
            raise Exception('Package not correctly installed: %s' % package.ident)

        # cleanup remove
        if os.path.exists(package.path):
            self.logger.info('pending remove %s', package.ident)
            tmp = package.path + '.tmp'
            shutil.move(package.path, tmp)
            self.logger.info('remove %s', package.ident)
            shutil.rmtree(tmp)

        self.load()
